"""This module implents helper functions for performance evaluation.
"""

import numpy as np
from ddn3 import tools


def scan_error_measure(t1_lst, t2_lst, comm_gt, diff_gt):
    """Estimate common and differential network error based on a list of estimations

    Let P be the number of features, and M be the number of estimates.

    Parameters
    ----------
    t1_lst : array_like
        Estimated precision matrix for condition 1 of shape (M, P, P).
    t2_lst : array_like
        Estimated precision matrix for condition 2 shape (M, P, P).
    comm_gt : ndarray
        Ground truth common network of shape (P, P).
    diff_gt
        Ground truth differential network of shape (P, P).

    Returns
    -------
    res_comm : ndarray
        Performance measures for the common network estimation. Shape (M, 5).
    res_diff : ndarray
        Performance measures for the differential network estimation. Shape (M, 5).

    """
    res_comm = np.zeros((len(t1_lst), 5))
    res_diff = np.zeros((len(t1_lst), 5))
    for i in range(len(t1_lst)):
        comm_est, diff_est = tools.get_common_diff_net_topo([t1_lst[i], t2_lst[i]])
        res_comm[i] = get_error_measure_two_theta(comm_est, comm_gt)
        res_diff[i] = get_error_measure_two_theta(diff_est, diff_gt)
    return res_comm, res_diff


def scan_erro_measure_dingo(comm_dingo, diff_dingo, comm_gt, diff_gt):
    """Estimate common and differential network error for DINGO.

    For DINGO, we apply different thresholds on the common network precision matrix, and differential score matrix to
    get the binary adjacency matrix.

    Let P be the number of features. Let M be the number thresholds.

    Parameters
    ----------
    t1_lst : array_like
        Estimated precision matrix for condition 1 of shape (P, P).
    t2_lst : array_like
        Estimated precision matrix for condition 2 shape (P, P).
    comm_gt : ndarray
        Ground truth common network of shape (P, P).
    diff_gt
        Ground truth differential network of shape (P, P).

    Returns
    -------
    res_comm_dingo : ndarray
        Performance measures for the common network estimation. Shape (M, 5).
    res_diff_dingo : ndarray
        Performance measures for the differential network estimation. Shape (M, 5).

    """
    thr_rg_comm = np.arange(0, 1, 0.01)
    res_comm_dingo = np.zeros((len(thr_rg_comm), 5))
    for i, thr in enumerate(thr_rg_comm):
        comm_est = tools.get_net_topo_from_mat(comm_dingo, thr=thr)
        res_comm_dingo[i] = get_error_measure_two_theta(comm_est, comm_gt)

    thr_rg_diff = np.arange(0, 50, 0.1)
    res_diff_dingo = np.zeros((len(thr_rg_diff), 5))
    for i, thr in enumerate(thr_rg_diff):
        diff_est = tools.get_net_topo_from_mat(diff_dingo, thr=thr)
        res_diff_dingo[i] = get_error_measure_two_theta(diff_est, diff_gt)
    return res_comm_dingo, res_diff_dingo


def get_error_measure_two_theta(net_est, net_gt):
    # See, e.g., https://en.wikipedia.org/wiki/Confusion_matrix for details
    n_node = len(net_est)
    n_edge = n_node * (n_node - 1) / 2

    P = np.sum(net_gt) / 2
    N = n_edge - P

    # number of true positive edges (TP) and false positive edges (FP)
    TP = np.sum((net_est > 0) * (net_gt > 0)) / 2
    FP = np.sum((net_est > 0) * (net_gt == 0)) / 2
    TP_FP = TP + FP

    # true positive rate, recall
    if P > 0:
        TPR = TP / P
    else:
        TPR = 0

    # false positive rate
    if N > 0:
        FPR = FP / N
    else:
        FPR = 0

    # precision
    if TP_FP > 0:
        PPV = TP / TP_FP
    else:
        PPV = 0

    return np.array([TP, FP, TPR, FPR, PPV])


def get_f1(recall, precision):
    """Calculate F1 score based on precision and recall"""
    recall1 = np.copy(recall)
    recall1[recall == 0] = 1e-8
    f1 = 2 * recall1 * precision / (recall1 + precision)
    f1[recall == 0] = 0
    f1[precision == 0] = 0
    return f1



"""The module for tuning hyperparameters in DDN

Find strategies are implemented.

- `cv_joint`: grid search CV for lambda1 and lambda2. This is time-consuming for larger data.
- `cv_sequential`: CV for lambda1 first, then use the determined lambda1 to do CV on lambda2.
- `cv_bai`:  CV for lambda1 first, then use the method in [1] to directly calculate lambda2.
- `mb_cv`: use theorem 3 in [2] to directly calculate lambda1, and use CV to get lambda2.
- `mb_bai`: use theorem 3 in [2] to directly calculate lambda1, and the method in [1] to get lambda2.

If the data set is not too large, `cv_joint` is a better choice.
If the data is larger, it is better to use the `cv_sequential` option.
If the data is really large, consider `cv_bai`.

A bette approach is to manually choose a set of lambda1 and select the one that leads to a reasonable network.
By utilizing the prior knowledge, it is more likely to obtain the network that is usable.

[1] "Learning structural changes of Gaussian graphical models in controlled experiments." UAI (2012).
[2] "High-dimensional graphs and variable selection with the lasso." Ann. Statist. (2006): 1436-1462.

"""

import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import norm
from ddn3 import ddn
from ddn3 import tools


class DDNParameterSearch:
    def __init__(
        self,
        dat1,
        dat2,
        lambda1_list=np.arange(0.05, 1.05, 0.05),
        lambda2_list=np.arange(0.025, 0.525, 0.025),
        n_cv=5,
        ratio_validation=0.2,
        alpha1=0.05,
        alpha2=0.01,
    ) -> None:
        """Initialize the parameter search.

        A number of parameters are set here. However, some of them are only used for certain parameter tuning methods.

        Parameters
        ----------
        dat1 : array_like
            The first data.
        dat2 : array_like
            The second data.
        lambda1_list : array_like
            A list of lambda1 values to search
        lambda2_list : array_like
            A list of lambda2 values to search
        n_cv : int
            The number of repeats. As we use repeated K-fold CV, `n_cv` can be as large as you like.
            Usually, a value of 10 or 20 is sufficient.
        ratio_validation : float
            The ratio of data that we used for validation. The remaining data is for training.
            As we use repeated CV, each time the training data is different.
        alpha1 : float
            The alpha used in theorem 3 of MB algorithm [2].
        alpha2 : float
            The alpha used in the method in [1]
        """
        # parameters
        self.dat1 = dat1
        self.dat2 = dat2
        self.l1_lst = lambda1_list
        self.l2_lst = lambda2_list
        self.n_cv = n_cv
        self.ratio_val = ratio_validation
        self.alpha1 = alpha1
        self.alpha2 = alpha2

        # derived and outputs
        self.n = int((dat1.shape[0] + dat2.shape[0]) / 2)
        self.p = dat1.shape[1]

    def fit(self, method="cv_sequential"):
        """Estimate lambda1 and lambda2.

        The validation error is defined as the ratio of the signal in the validation set that cannot be
         explained by the network estimated in the training set.

        Parameters
        ----------
        method : str
            The method used for estimation.

        Returns
        -------
        out : tuple
            The validation error, estimated lambda1, and lambda2

        """
        if method == "cv_joint":
            out = self.run_cv_joint()
        elif method == "cv_sequential":
            out = self.run_cv_sequential()
        elif method == "cv_bai":
            out = self.run_cv_bai()
        elif method == "mb_cv":
            out = self.run_mb_cv()
        elif method == "mb_bai":
            out = self.run_mb_bai()
        else:
            out = self.run_cv_bai()
        return out

    def run_cv_joint(self):
        """Estimation lambda1 and lambda2 with grid search CV"""
        # can be slow
        val_err, _, _ = cv_two_lambda(
            self.dat1,
            self.dat2,
            n_cv=self.n_cv,
            ratio_val=self.ratio_val,
            lambda1_lst=self.l1_lst,
            lambda2_lst=self.l2_lst,
        )
        l1_est, l2_est = get_lambdas_one_se_2d(val_err, self.l1_lst, self.l2_lst)
        return val_err, l1_est, l2_est

    def run_cv_sequential(self):
        """Use CV to estimate lambda1, then use CV for lambda2"""
        val_err1, _, _ = cv_two_lambda(
            self.dat1,
            self.dat2,
            lambda1_lst=self.l1_lst,
            lambda2_lst=[0.0],
            n_cv=self.n_cv,
            ratio_val=self.ratio_val,
        )
        val_err1 = np.squeeze(val_err1)
        l1_est = get_lambda_one_se_1d(val_err1, self.l1_lst)

        val_err2, _, _ = cv_two_lambda(
            self.dat1,
            self.dat2,
            lambda1_lst=[l1_est],
            lambda2_lst=self.l2_lst,
            n_cv=self.n_cv,
            ratio_val=self.ratio_val,
        )
        val_err2 = np.squeeze(val_err2)
        l2_est = get_lambda_one_se_1d(val_err2, self.l2_lst)

        return [val_err1, val_err2], l1_est, l2_est

    def run_cv_bai(self):
        """Use CV for lambda1, use the method in [1] for lambda2"""
        val_err, _, _ = cv_two_lambda(
            self.dat1,
            self.dat2,
            lambda1_lst=self.l1_lst,
            lambda2_lst=[0.0],
            n_cv=self.n_cv,
            ratio_val=self.ratio_val,
        )
        val_err = np.squeeze(val_err)
        l1_est = get_lambda_one_se_1d(val_err, self.l1_lst)
        l2_est = get_lambda2_bai(self.dat1, self.dat2, alpha=self.alpha2)

        return val_err, l1_est, l2_est

    def run_mb_cv(self):
        """Use the theorem 3 in [2] for lambda1, and use CV for lambda2"""
        # mb not reliable with small sample size
        l1_est = get_lambda1_mb(self.alpha1, self.n, self.p)
        val_err, _, _ = cv_two_lambda(
            self.dat1,
            self.dat2,
            lambda1_lst=[l1_est],
            lambda2_lst=self.l2_lst,
            n_cv=self.n_cv,
            ratio_val=self.ratio_val,
        )
        val_err = np.squeeze(val_err)
        l2_est = get_lambda_one_se_1d(val_err, self.l2_lst)

        return val_err, l1_est, l2_est

    def run_mb_bai(self):
        """Use theorem 3 in [2] for lambda1, and use method in [1] for lambda2"""
        # mb not reliable with small sample size
        l1_est = get_lambda1_mb(self.alpha1, self.n, self.p, mthd=0)
        l2_est = get_lambda2_bai(self.dat1, self.dat2, alpha=self.alpha2)
        return [], l1_est, l2_est


def get_lambda1_mb(alpha, n, p, mthd=0):
    """Estimate lambda1 using theorem 3 in [2]

    Parameters
    ----------
    alpha : float
        The parameter controlling false positives
    n : int
        The number of samples
    p : int
        The number of features
    mthd : int
        Use 0 for methods in [2], with a correction term of 2.
        Use 1 for methods used in https://pubmed.ncbi.nlm.nih.gov/25273109/

    Returns
    -------
    lmb1 : float
        The estimated lambda1

    """
    lmb1 = 0.5
    if mthd == 0:
        # The MB paper do not have 1/2 factor for data term
        # To make things consistent, we further divide by 2
        lmb1 = 2 / np.sqrt(n) * norm.ppf(1 - alpha / (2 * p * n * n)) / 2
        # lmb1 = 2 / np.sqrt(n) * norm.ppf(1 - alpha / (2 * p * n * n))
    if mthd == 1:
        # NOTE: this is from KDDN Java code, which is not the same as MB paper
        lmb1 = 2 / n * norm.ppf(1 - alpha / (2 * p * n * n))
    return lmb1


def get_lambda2_bai(
    x1,
    x2,
    alpha=0.01,
):
    """Estimate lambda2 using the method in [1]

    Parameters
    ----------
    x1 : array_like
        The data for condition 1
    x2 : array_like
        The data for condition 2
    alpha : float
        The parameter controlling false positives

    Returns
    -------
    lmb2 : float
        Estimated labmda2

    """
    x1 = tools.standardize_data(x1)
    x2 = tools.standardize_data(x2)

    n1 = x1.shape[0]
    n2 = x2.shape[0]
    N = (n1 + n2) / 2
    p = x1.shape[1]

    s = norm.ppf(1 - alpha / 2) * np.sqrt(2 / (N - 3))

    # pd = (x1.T @ x1) * (x2.T @ x2)
    pd = (x1.T @ x1 / n1) * (x2.T @ x2 / n2)
    pd[np.arange(p), np.arange(p)] = 0
    rho1rho2 = np.sum(pd) / p / (p - 1)

    lmb2 = (np.exp(2 * s) - 1) / (np.exp(2 * s) + 1) / 2 * (1 - rho1rho2)
    print("Avg rho is ", rho1rho2)
    print("lambda2 is ", lmb2)

    return lmb2


def get_lambda_one_se_1d(val_err, lambda_lst):
    """Choose a lambda from a list of lambda values using the one standard error rule

    let K be the number of CV repeats, L the number of lambda values.

    Parameters
    ----------
    val_err : array_like
        Validation errors corresponding to the list of lambda. Shape K by L.
    lambda_lst : array_like
        The lambda values. Shape L.

    Returns
    -------
    float
        Chosen lambda value

    """
    val_err_mean = np.mean(val_err, axis=0)
    val_err_std = np.std(val_err, axis=0)
    n_cv = val_err.shape[0]

    idx = np.argmin(val_err_mean)
    val_err_mean[:idx] = -10000

    cut_thr = val_err_mean[idx] + val_err_std[idx] / np.sqrt(n_cv)  # standard error
    if np.max(val_err_mean) < cut_thr:
        print("One SE rule failed.")
        return lambda_lst[-1]
    else:
        # idx_thr = np.max(np.where(val_err_mean <= cut_thr)[0])
        idx_thr = np.min(np.where(val_err_mean >= cut_thr)[0])
        return lambda_lst[idx_thr]


def get_lambdas_one_se_2d(val_err, lambda1_lst, lambda2_lst):
    """Choose a lambda from a list of lambda values using the one standard error rule

    After finding the minimum value, we find all the combinations of lambda1 and lambda2 values that are at least one
    standard error larger than this minimum value.
    From these combinations, we choose the one whose lambda1 and lambda2 values are closest to those of the minimum.

    let K be the number of CV repeats, L1 the number of lambda1 values, L2 the number of lambda2 values.

    Parameters
    ----------
    val_err : array_like
        Validation errors corresponding to the list of lambda. Shape K by L1 by L2.
    lambda1_lst : array_like
        The lambda1 values, shape L1.
    lambda1_lst : array_like
        The lambda2 values, shape L2.

    Returns
    -------
    float
        Chosen lambda value

    """
    gap1 = np.abs(lambda1_lst[-1] - lambda1_lst[0]) / len(lambda1_lst)
    gap2 = np.abs(lambda2_lst[-1] - lambda2_lst[0]) / len(lambda2_lst)
    lmb1_scale = gap1 / gap2

    val_err_mean = np.mean(val_err, axis=0)
    val_err_std = np.std(val_err, axis=0)
    n_cv = val_err.shape[0]

    idx1, idx2 = np.unravel_index(val_err_mean.argmin(), val_err_mean.shape)
    print(idx1, idx2)
    print("l1_org, l2_org ", lambda1_lst[idx1], lambda2_lst[idx2])
    msk = np.zeros_like(val_err_mean)
    msk[idx1:, idx2:] = 1.0
    se = val_err_std[idx1, idx2] / np.sqrt(n_cv)
    z = (val_err_mean - np.min(val_err_mean)) / se
    z = z * msk

    if np.max(z) == 0:
        print("One SE rule failed.")
        return np.max(lambda1_lst), np.max(lambda2_lst)
    else:
        m1, m2 = val_err_mean.shape
        cord1, cord2 = np.meshgrid(np.arange(m1), np.arange(m2), indexing="ij")
        d = (cord1 - idx1) ** 2 * lmb1_scale + (cord2 - idx2) ** 2
        d[z < 1] = 100000
        idx1a, idx2a = np.unravel_index(d.argmin(), d.shape)
        print(idx1a, idx2a)
        print("l1, l2 ", lambda1_lst[idx1a], lambda2_lst[idx2a])

        return lambda1_lst[idx1a], lambda2_lst[idx2a]


def calculate_regression(data, topo_est):
    """Linear regression

    For each variable, use all its neighbors as  predictors and find the regression coefficients.

    This is an example of regression operation

    >>> x = np.array([[-1,-1,1,1.0], [1,1,-1,-1]]).T
    >>> y = np.array([1,1,-1,-1.0])
    >>> out = np.linalg.lstsq(x, y, rcond=None)
    >>> out[0]

    Parameters
    ----------
    data : array_like
        All data
    topo_est : array_like
        Estimated adjacency matrix

    Returns
    -------

    """
    n_fea = data.shape[1]
    g_asso = np.eye(n_fea, dtype=np.double)
    for i in range(n_fea):
        pred_idx = np.where(topo_est[i] > 0)[0]
        if len(pred_idx) == 0:
            continue
        y = data[:, i]
        x = data[:, pred_idx]
        out = np.linalg.lstsq(x, y, rcond=None)
        g_asso[i, pred_idx] = out[0]
    return g_asso


def cv_two_lambda(
    dat1,
    dat2,
    n_cv=5,
    ratio_val=0.2,
    lambda1_lst=np.arange(0.05, 1.05, 0.05),
    lambda2_lst=np.arange(0.025, 0.525, 0.025),
):
    """Cross validation by grid search lambda1 and lambda2

    To estimate the validation error, we estimate the coefficient of each node on the training set based on the
    estimated network topology. Then for each node in the validation set, we try to use its neighbors to explain the
    signal in that node. The portion of unexplained signal in all nodes is defined as the validation error.

    Let K be the number of CV repeats, L1 the number of lambda1 values, L2 the number of lambda2 values.

    Parameters
    ----------
    dat1 : array_like
        Data for condition 1
    dat2 : array_like
        Data for condition 1
    n_cv : int
        Number of repeats. Can be as large as you like, as we re-sample each time.
    ratio_val : float
        Ratio of data for validation. The remaining is used for training.
    lambda1_lst : array_like
        Values of lambda1 for searching
    lambda2_lst : array_like
        Values of lambda2 for searching

    Returns
    -------
    val_err : array_like
        The validation error for each lambda1 and lambda2 combination. Shape K by L1 by L2
    lambda1_lst : array_like
    lambda2_lst : array_like

    """
    val_err = np.zeros((n_cv, len(lambda1_lst), len(lambda2_lst)))
    n_node = dat1.shape[1]
    n1 = dat1.shape[0]
    n2 = dat2.shape[0]
    n1_val = int(n1 * ratio_val)
    n1_train = n1 - n1_val
    n2_val = int(n2 * ratio_val)
    n2_train = n2 - n2_val

    mthd = "resi"
    # if len(lambda2_lst) == 1 and lambda2_lst[0] == 0:
    #     mthd = 'strongrule'
    print(mthd)

    for n in range(n_cv):
        print("Repeat ============", n)
        msk1 = np.zeros(n1)
        msk1[np.random.choice(n1, n1_train, replace=False)] = 1
        msk2 = np.zeros(n2)
        msk2[np.random.choice(n2, n2_train, replace=False)] = 1
        g1_train = tools.standardize_data(dat1[msk1 > 0])
        g1_val = tools.standardize_data(dat1[msk1 == 0])
        g2_train = tools.standardize_data(dat2[msk2 > 0])
        g2_val = tools.standardize_data(dat2[msk2 == 0])

        for i, lambda1 in enumerate(lambda1_lst):
            # print(n, i)
            for j, lambda2 in enumerate(lambda2_lst):
                g_beta_est = ddn.ddn(
                    g1_train,
                    g2_train,
                    lambda1=lambda1,
                    lambda2=lambda2,
                    mthd=mthd,
                )
                g1_net_est = tools.get_net_topo_from_mat(g_beta_est[0])
                g2_net_est = tools.get_net_topo_from_mat(g_beta_est[1])
                g1_coef = calculate_regression(g1_train, g1_net_est)
                g1_coef[np.arange(n_node), np.arange(n_node)] = 0
                g2_coef = calculate_regression(g2_train, g2_net_est)
                g2_coef[np.arange(n_node), np.arange(n_node)] = 0
                rec_ratio1 = np.linalg.norm(
                    g1_val @ g1_coef.T - g1_val
                ) / np.linalg.norm(g1_val)
                rec_ratio2 = np.linalg.norm(
                    g2_val @ g2_coef.T - g2_val
                ) / np.linalg.norm(g2_val)
                val_err[n, i, j] = (rec_ratio1 + rec_ratio2) / 2

    return val_err, lambda1_lst, lambda2_lst


def plot_error_1d(val_err, lambda_lst=(), ymin=None, ymax=None):
    """Plot the curve of cross validation with one lambda"""
    val_err_mean = np.mean(val_err, axis=0)
    val_err_std = np.std(val_err, axis=0)
    n_cv = val_err.shape[0]
    if not ymin:
        ymin = max(np.min(val_err_mean), 0.0)
    if not ymax:
        ymax = min(np.max(val_err_mean), 1.0)
    gap1 = (ymax - ymin) / 10
    fig, ax = plt.subplots()
    ax.errorbar(
        lambda_lst,
        val_err_mean,
        yerr=val_err_std / np.sqrt(n_cv),
        ecolor="red",
        elinewidth=0.5,
    )
    ax.set_ylim([ymin - gap1, ymax + gap1])


def plot_error_2d(val_err, cmin=None, cmax=None):
    """Plot the image of cross validation with both lambda1 and lambda2"""
    val_err_mean = np.mean(val_err, axis=0)
    val_err_std = np.std(val_err, axis=0)
    n_cv = val_err.shape[0]
    if not cmin:
        cmin = np.min(val_err_mean)
    if not cmax:
        cmax = np.max(val_err_mean)
    fig, ax = plt.subplots()
    pos = ax.imshow(val_err_mean, origin="lower", vmin=cmin, vmax=cmax)
    color_bar = fig.colorbar(pos, ax=ax)
    ax.set_xlabel("$\lambda_2$")
    ax.set_ylabel("$\lambda_1$")



"""Wrapper functions for calling the BCD algorithms
"""

import numpy as np
from ddn3 import tools
from ddn3 import bcd, strong_rule


def run_org(
    g1_data,
    g2_data,
    node,
    lambda1,
    lambda2,
    beta1_in,
    beta2_in,
    threshold=1e-6,
    use_warm=False,
):
    """The wrapper that calls the DDN 2.0 BCD algorithm

    Denote P be the number features. N1 be the sample size for condition 1, and N2 for condition 2.

    Parameters
    ----------
    g1_data : array_like, shape N1 by P
        The data from condition 1
    g2_data : array_like, shape N2 by P
        The data from condition 2
    node : int
        Index of the current node that serve as the response variable.
    lambda1 : float
        DDN parameter lambda1.
    lambda2 : float
        DDN parameter lambda2.
    beta1_in : array_like, length P
        Initial beta for condition 1. If initialization is not needed, use an array of all zeros.
    beta2_in : array_like, length P
        Initial beta for condition 2. If initialization is not needed, use an array of all zeros.
    threshold : float
        Convergence threshold.
    use_warm : bool, optional
        Not used. To use warm start, simply input `beta1_in` and `beta2_in` from previous hyperparameters.

    Returns
    -------
    beta1 : ndarray, length P
        Estimated beta for `node` in condition 1.
    beta2 : ndarray, length P
        Estimated beta for `node` in condition 2.

    """
    beta_in = np.concatenate(
        (
            beta1_in[:node],
            beta1_in[node + 1 :],
            beta2_in[:node],
            beta2_in[node + 1 :],
        )
    )

    N_NODE = g1_data.shape[1]
    n1, n2 = g1_data.shape[0], g2_data.shape[0]

    y = tools.concatenate_data(g1_data[:, node], g2_data[:, node], method="row")
    node_fea = [i for i in range(N_NODE) if i != node]
    X = tools.concatenate_data(
        g1_data[:, node_fea], g2_data[:, node_fea], method="diag"
    )
    beta, r, betaerr = bcd.bcd_org(beta_in, y, X, lambda1, lambda2, n1, n2, threshold)

    # reindex the features
    beta1 = list(beta[0:node]) + [0] + list(beta[node : N_NODE - 1])
    beta1 = np.array(beta1)
    beta2 = (
        list(beta[N_NODE - 1 : node + N_NODE - 1])
        + [0]
        + list(beta[node + N_NODE - 1 : 2 * N_NODE - 2])
    )
    beta2 = np.array(beta2)
    return beta1, beta2


def run_resi(
    g1_data,
    g2_data,
    node,
    lambda1,
    lambda2,
    beta1_in,
    beta2_in,
    threshold,
    use_warm=False,
):
    """The wrapper that calls the DDN 3.0 residual update algorithm

    If warm start is used, the residual signal is updated accordingly to previously estimated beta.
    This takes some extra time, and may make warm start less appealing.

    Denote P be the number features. N1 be the sample size for condition 1, and N2 for condition 2.

    Parameters
    ----------
    g1_data : array_like, shape N1 by P
        The data from condition 1
    g2_data : array_like, shape N2 by P
        The data from condition 2
    node : int
        Index of the current node that serve as the response variable.
    lambda1 : float
        DDN parameter lambda1.
    lambda2 : float
        DDN parameter lambda2.
    beta1_in : array_like, length P
        Initial beta for condition 1. If initialization is not needed, use an array of all zeros.
    beta2_in : array_like, length P
        Initial beta for condition 2. If initialization is not needed, use an array of all zeros.
    threshold : float
        Convergence threshold.
    use_warm : bool, optional
        If true, use warm start.

    Returns
    -------
    beta1 : ndarray, length P
        Estimated beta for `node` in condition 1.
    beta2 : ndarray, length P
        Estimated beta for `node` in condition 2.

    """
    beta_in = np.concatenate((beta1_in, beta2_in))
    if use_warm:
        beta1_in_x = np.copy(beta1_in)
        beta2_in_x = np.copy(beta2_in)
        beta1_in_x[node] = 0
        beta2_in_x[node] = 0
        y1_resi = g1_data[:, node] - np.dot(g1_data, beta1_in_x)
        y2_resi = g2_data[:, node] - np.dot(g2_data, beta2_in_x)
    else:
        y1_resi = np.copy(g1_data[:, node])
        y2_resi = np.copy(g2_data[:, node])

    N_NODE = g1_data.shape[1]

    # beta = bcd_residual(g1_data, g2_data, node, lambda1, lambda2, threshold)
    beta, r, betaerr = bcd.bcd_residual(
        beta_in,
        g1_data,
        g2_data,
        y1_resi,
        y2_resi,
        node,
        lambda1,
        lambda2,
        threshold,
    )
    # n_iter.append(r)
    beta1 = np.array(beta[:N_NODE])
    beta2 = np.array(beta[N_NODE:])

    return beta1, beta2


def run_corr(
    corr_matrix_1,
    corr_matrix_2,
    node,
    lambda1,
    lambda2,
    beta1_in,
    beta2_in,
    threshold,
    use_warm=False,
):
    """The wrapper that calls the DDN 3.0 correlation matrix update algorithm

    Parameters
    ----------
    corr_matrix_1 : ndarray
        Input correlation matrix for condition 1.
    corr_matrix_2 : ndarray
        Input correlation matrix for condition 2.
    node : int
        Index of the current node that serve as the response variable.
    lambda1 : float
        DDN parameter lambda1.
    lambda2 : float
        DDN parameter lambda2.
    beta1_in : array_like, length P
        Initial beta for condition 1. If initialization is not needed, use an array of all zeros.
    beta2_in : array_like, length P
        Initial beta for condition 2. If initialization is not needed, use an array of all zeros.
    threshold : float
        Convergence threshold.
    use_warm : bool, optional
        If true, use warm start.

    Returns
    -------
    beta1 : ndarray, length P
        Estimated beta for `node` in condition 1.
    beta2 : ndarray, length P
        Estimated beta for `node` in condition 2.
    """
    beta_in = np.concatenate((beta1_in, beta2_in))

    beta, r, betaerr = bcd.bcd_corr(
        beta_in,
        node,
        lambda1,
        lambda2,
        corr_matrix_1,
        corr_matrix_2,
        threshold=threshold,
        max_iter=100000,
    )

    n_node = len(beta1_in)
    beta1 = np.array(beta[:n_node])
    beta2 = np.array(beta[n_node:])

    return beta1, beta2


def run_strongrule(
    g1_data,
    g2_data,
    node,
    lambda1,
    lambda2,
    beta1_in,
    beta2_in,
    threshold,
    use_warm=False,
):
    """The wrapper that calls the DDN 3.0 strong rule strategy.

    In this algorithm, `lambda2` is kept as zero. Then DDN becomes two separate lasso problem.
    The strong rule can then be utilized to eliminate part of the predictor variables for each `lambda1`.
    The lasso problem is also solved by the `lasso` function of scikit-learn.

    Warm start is generally preferred when using this strategy.

    Note that empirically this strategy is not fast.

    Denote P be the number features. N1 be the sample size for condition 1, and N2 for condition 2.

    Parameters
    ----------
    g1_data : array_like, shape N1 by P
        The data from condition 1
    g2_data : array_like, shape N2 by P
        The data from condition 2
    node : int
        Index of the current node that serve as the response variable.
    lambda1 : float
        DDN parameter lambda1.
    lambda2 : float
        Not used. Must be 0.
    beta1_in : array_like, length P
        Initial beta for condition 1. If initialization is not needed, use an array of all zeros.
    beta2_in : array_like, length P
        Initial beta for condition 2. If initialization is not needed, use an array of all zeros.
    threshold : float
        Convergence threshold.
    use_warm : bool, optional
        Use warm start or not.

    Returns
    -------
    beta1 : ndarray, length P
        Estimated beta for `node` in condition 1.
    beta2 : ndarray, length P
        Estimated beta for `node` in condition 2.

    """
    beta1_in_x = (np.concatenate([beta1_in[:node], beta1_in[node + 1 :]]),)
    beta2_in_x = (np.concatenate([beta2_in[:node], beta2_in[node + 1 :]]),)
    y1 = g1_data[:, node]
    y2 = g2_data[:, node]

    N_NODE = g1_data.shape[1]

    # choose other genes as feature
    idx = [i for i in range(N_NODE) if i != node]
    X1 = g1_data[:, idx]
    X2 = g2_data[:, idx]

    # perform bcd algorithm
    beta1 = strong_rule.lasso(
        y1,
        X1,
        lambda1,
        beta1_in_x,
        tol=threshold,
        use_strong_rule=True,
        use_warm=use_warm,
    )
    beta2 = strong_rule.lasso(
        y2,
        X2,
        lambda1,
        beta2_in_x,
        tol=threshold,
        use_strong_rule=True,
        use_warm=use_warm,
    )

    # reindex the features
    beta1 = list(beta1[0:node]) + [0] + list(beta1[node:])
    beta1 = np.array(beta1)
    beta2 = list(beta2[0:node]) + [0] + list(beta2[node:])
    beta2 = np.array(beta2)

    return beta1, beta2



"""Utility functions of DDN
"""

import numpy as np


def standardize_data(data, scaler="std", zero_mean_already=False):
    """Standadize each column of the input data

    Parameters
    ----------
    data : ndarray
        Input data. Each column is a feature.
    scaler : str, optional
        Method to calculate the scale of each feature, by default "std"
    zero_mean_already : bool, optional
        The data is already with zero mean, by default False

    Returns
    -------
    standard_data : ndarray
        Standardized data
    """

    # sample standardization : z = (x - u) / s

    standard_data = np.zeros(data.shape)
    for i in range(data.shape[1]):
        # mean value
        u = np.mean(data[:, i]) if not zero_mean_already else 0

        if scaler == "std":
            # standard deviation
            s = np.std(data[:, i])
        elif scaler == "rms":
            # root-mean-square
            s = np.sqrt(np.mean(np.square(data[:, i])))
        else:
            s = 1

        standard_data[:, i] = (data[:, i] - u) / s

    return standard_data


def gen_mv(cov_mat, n_sample):
    """Generate zero mean multivariate data and standardize it.

    Parameters
    ----------
    cov_mat : ndarray
        Covariance matrix
    n_sample : int
        Sample size

    Returns
    -------
    ndarray
        Generated samples
    """
    x = np.random.multivariate_normal(np.zeros(len(cov_mat)), cov_mat, n_sample)
    return standardize_data(x, scaler="std")


def concatenate_data(controldata, casedata, method="diag"):
    """Several method to concatenate two arrays"""
    if method == "row":
        return np.concatenate((controldata, casedata), axis=0)
    elif method == "col":
        return np.concatenate((controldata, casedata), axis=1)
    elif method == "diag":
        return np.concatenate(
            (
                np.concatenate((controldata, casedata * 0), axis=0),
                np.concatenate((controldata * 0, casedata), axis=0),
            ),
            axis=1,
        )
    else:
        return []


def get_net_topo_from_mat(mat_prec, thr=1e-4):
    """Threshold the input matrxi to get adjacency matrix

    The input is also made symmetry.
    As we do not use self loop, the diagonal elements are always 0.

    Parameters
    ----------
    mat_prec : ndarray
        Input matrix. It could be a precision matrix, or a coefficient matrix.
    thr : float, optional
        Threshold to remove too small items in `mat_prec`, by default 1e-4

    Returns
    -------
    ndarray
        Adjacency matrix
    """
    N = len(mat_prec)
    x = np.copy(mat_prec)
    x[np.arange(N), np.arange(N)] = 0.0
    x = 1.0 * (np.abs(x) > thr)
    x = 1.0 * ((x + x.T) > 0)
    return x


def get_common_diff_net_topo(g_beta, thr=1e-4):
    """Calcualte common and differential network from the output of DDN

    g_beta[0] is the coefficient matrxi for condition 1
    g_beta[0] is the coefficient matrxi for condition 2
    """
    g1 = get_net_topo_from_mat(g_beta[0], thr=thr)
    g2 = get_net_topo_from_mat(g_beta[1], thr=thr)
    g_net_comm = 1.0 * ((g1 + g2) == 2)
    g_net_dif = 1.0 * (g1 != g2)
    return g_net_comm, g_net_dif


def ddn_obj_fun(y, X, lambda1, lambda2, n1, n2, beta):
    """Objective function for DDN, assuming two conditions having the same sample size"""
    p = X.shape[1] // 2
    beta1 = beta[:p]
    beta2 = beta[p:]
    d0 = y - X @ beta
    res = (
        np.sum(d0 * d0) / n1 / 2
        + np.sum(np.abs(beta)) * lambda1
        + np.sum(np.abs(beta1 - beta2)) * lambda2
    )
    return res



"""Utility function for simple simulation

These simulations are designed to illustrate the usage DDN, and not for formal evaluation of performance.
Only the pair graph is supported.

It also contains some helper functions for other simulation functions.

"""

import numpy as np
from ddn3 import tools


def create_pair_graph(
    n_node=40,
    corr=0.75,
    n_shuf=3,
):
    """Generate precision matrix for pair graph.

    For graph with N nodes, there are N/2 edges, and each node has degree one.

    Parameters
    ----------
    n_node : int, optional
        The number of nodes, by default 40
    corr : float, optional
        The value of precision matrix between two neighboring nodes, by default 0.75
    n_shuf : int, optional
        The number of edges to shuffle, by default 3

    Returns
    -------
    g1_prec : ndarray
        Precision matrix for condition 1, shape (N, N)
    g2_prec : ndarray
        Precision matrix for condition 2, shape (N, N)
    """
    n_blk = int(n_node / 2)
    edge1 = np.zeros((n_blk, 2))
    edge1[:, 0] = np.arange(n_blk)
    edge1[:, 1] = np.arange(n_blk) + n_blk

    edge2 = np.copy(edge1)
    idx_shuf = np.arange(n_blk, n_blk + n_shuf)
    edge2[: len(idx_shuf), 1] = np.concatenate([idx_shuf[[-1]], idx_shuf[:-1]])

    xx = np.array([[1, -corr], [-corr, 1]])
    # print(np.linalg.eig(xx))

    g1_prec = np.zeros((n_node, n_node))
    for e in edge1:
        e = e.astype(int)
        g1_prec[np.ix_([e[0], e[1]], [e[0], e[1]])] = xx

    g2_prec = np.zeros((n_node, n_node))
    for e in edge2:
        e = e.astype(int)
        g2_prec[np.ix_([e[0], e[1]], [e[0], e[1]])] = xx

    return g1_prec, g2_prec


def prep_sim_from_two_omega(omega1, omega2):
    """Get covariance matrices and adjacency matrices from two precisions matrices

    Parameters
    ----------
    omega1 : array_like
        The precision matrix for condition 1
    omega2 : array_like
        The precision matrix for condition 2

    Returns
    -------
    g1_cov : ndarray
        Covariance matrix for condition 1
    g2_cov : ndarray
        Covariance matrix for condition 2
    comm_gt : ndarray
        Adjacency matrix of common network
    diff_gt : ndarray
        Adjacency matrix of differential network
    """
    g1_cov, _ = create_cov_prec_mat(omega1)
    g2_cov, _ = create_cov_prec_mat(omega2)
    comm_gt, diff_gt = tools.get_common_diff_net_topo([omega1, omega2])
    return g1_cov, g2_cov, comm_gt, diff_gt


def gen_sample_two_conditions(g1_cov, g2_cov, n1, n2):
    """Generate multivariante normal samples for two conditions

    Let P be the number of features.

    Parameters
    ----------
    g1_cov : array_like
        Covariance matrix for condition 1
    g2_cov : array_like
        Covariance matrix for condition 2
    n1 : int
        Number of samples for condition 1
    n2 : int
        Number of samples for condition 2

    Returns
    -------
    dat1 : ndarray
        Generated samples for condition 1. Shape (n1, P)
    dat2 : ndarray
        Generated samples for condition 2. Shape (n2, P)
    """
    dat1 = tools.gen_mv(g1_cov, n1)
    dat2 = tools.gen_mv(g2_cov, n2)
    return dat1, dat2


def create_cov_prec_mat(prec_mat_in):
    """Create covariance from temporary precision matrix

    Each variable now have unit variance. 
    We also provide the corresponding precision matrix.

    We follow [Peng 2009] and do not use the d_ij term as the JGL paper.

    Parameters
    ----------
    prec_mat_in : ndarray
        Input precision matrix

    Returns
    -------
    cov_mat : ndarray
        Modified covariance matrix
    prec_mat : ndarray
        Corresponding precision matrix
    """
    cov_mat_temp = np.linalg.inv(prec_mat_in)
    d_sqrt = np.sqrt(np.diag(1 / np.diagonal(cov_mat_temp)))
    cov_mat = d_sqrt @ cov_mat_temp @ d_sqrt
    prec_mat = np.linalg.inv(cov_mat)

    return cov_mat, prec_mat


def simple_data():
    """Generate example data for the tutorial"""
    n_node = 40
    n_sample1 = 100
    n_sample2 = 100
    n_shuf = 5
    g1_prec, g2_prec = create_pair_graph(n_node=n_node, corr=0.75, n_shuf=n_shuf)
    gene_names = [f"Gene{i}" for i in range(n_node)]
    g1_cov, _ = create_cov_prec_mat(g1_prec)
    g2_cov, _ = create_cov_prec_mat(g2_prec)
    dat1 = tools.gen_mv(g1_cov, n_sample1)
    dat2 = tools.gen_mv(g2_cov, n_sample2)

    return dat1, dat2, gene_names



"""The convenience wrappers for running DDN
"""

from ddn3 import ddn, tools_export
from ddn3.simulation import simple_data


def ddn_pipeline(dat1, dat2, gene_names, lambda1=0.3, lambda2=0.1):
    """A wrapper around DDN

    Reads two data sets and a list of genes, then runs DDN and reports common and differential networks.

    Parameters
    ----------
    dat1 : array_like
        The first data for DDN
    dat2 : array_like
        The second data for DDN
    gene_names : list of str
        List of gene names for reporting and visualization
    lambda1 : float, optional
        The parameter controlling the overall sparsity. Usually should be between 0 and 1
    lambda2 : float, optional
        The parameter controlling the differences between two networks. Usually should be between 0 and 0.3

    Returns
    -------
    comm_edge : pandas.DataFrame
        The common edges between the two networks
    diff_edge : pandas.DataFrame
        The differential edges between the two networks
    node_non_isolated : list of str
        List of nodes that have at least one neighboring nodes

    """
    omega1, omega2 = ddn.ddn(dat1, dat2, lambda1, lambda2)
    (
        comm_edge,
        _,
        _,
        diff_edge,
        node_non_isolated,
    ) = tools_export.get_diff_comm_net_for_plot(omega1, omega2, gene_names)
    return comm_edge, diff_edge, node_non_isolated



"""Visualization functions of DDN

These functions are quite basic. 
For advanced plotting of networks, consider using specialized tools.
"""

import numpy as np
import scipy as sp
from scipy.spatial.distance import pdist
import networkx as nx
import matplotlib.pyplot as plt


def draw_network_for_ddn(
    edges_df,
    nodes_to_draw,
    fig_size=18,
    font_size_scale=1,
    node_size_scale=2,
    part_number=1,
    nodes_type=None,
    labels=None,
    mode="common",
    export_pdf=True,
    pdf_name="",
):
    """Draw networks for DDN

    By default, we draw all nodes to a circle.
    It is possible for names of features to overlap with each other,
    especially at the top or bottom of the circle.
    If this happens, consider setting a smaller `font_size_scale`.

    For common network, if two nodes in an edge have same type, draw grey line.
    If  two nodes in an edge have different type, draw green line.

    For differential network, if an edge comes from condition 1, draw blue line.
    If an edge comes from condition 2, draw red line.

    We support drawing nodes in one or two types (like Gene vs. TF).
    If two types of nodes are present, we draw two ellipses.
    See `part_number` and `nodes_type` parameters.

    Parameters
    ----------
    edges_df : pandas.DataFrame
        Edge information. 
        First two columns for the two feature names. 
        Third column for edge type (common=0, diff1=1, diff2=2)
        Fourth column for weight.
    nodes_to_draw : list of str
        Name of nodes to draw
    fig_size : int, optional
        Size of figure, by default 18
        We draw in a square figure, so width=height=`fig_size`.
    font_size_scale : int, optional
        Scale of fonts, by default 1
        To make the font larger, set larger value.
        The default value is scaled according to the node number.
    node_size_scale : int, optional
        Scale of node sizes, by default 2
    part_number : int, optional
        Number of ellipse to draw, by default 1
        If set to 1, all nodes are in the same condition, draw a single circle.
        If set to 2, nodes are in two conditions (like Gene and TF), we draw two ellipses.
        If set to 2, `nodes_type` is needed.
    nodes_type : None or dict, optional
        Node type (e.g., Gene=0, TF=1), by default None
    labels : dict
        Alternative (e.g., simplfied) names for nodes
    mode : str, optional
        Draw common graph or differential graph, by default "common"
    export_pdf : bool, optional
        Set to true want to export the graph as a PDF file, by default True
    pdf_name : str, optional
        Name of the PDF file to export, by default ""

    Returns
    -------
    nx.Graph
        Graph object
    """
    # create networkx graph
    G = _create_nx_graph(
        edges_df,
        nodes_to_draw,
        mode=mode,
        nodes_type=nodes_type,
    )

    # nodes positions
    # must provide nodes_type dictionary for two parts
    if part_number == 1:
        pos, d_min = _get_pos_one_part(nodes_to_draw)
    elif part_number == 2:
        pos, d_min = _get_pos_two_parts(nodes_to_draw, nodes_type)
    else:
        raise ("Not implemented")

    # plot the network
    _plot_network_helper(
        G,
        pos,
        d_min=d_min,
        labels=labels,
        fig_size=fig_size,
        font_size_scale=font_size_scale,
        node_size_scale=node_size_scale,
    )

    # export figure
    if export_pdf:
        plt.savefig(f"{pdf_name}_{mode}.pdf", format="pdf", bbox_inches="tight")

    return G


def _add_node_to_a_circle(pos, nodes, cen, rad, angles):
    """Find positions in an ellipse, and calculate the minimum distances between points

    Parameters
    ----------
    pos : dict
        For saving the position results for NetworkX
    nodes : list of str
        Name of nodes
    cen : array_like
        Central position of this circle, shape (2,)
    rad : array_like
        Length of two axes of the ellipse, shape (2,)
    angles : array_like
        Angles of the points

    Returns
    -------
    float
        Minimum distances between points
    """
    pos_lst = []
    for i, node in enumerate(nodes):
        theta = angles[i]
        pos0 = np.array(
            [cen[0] + np.cos(theta) * rad[0], cen[1] + np.sin(theta) * rad[1]]
        )
        pos[node] = pos0
        pos_lst.append(pos0)
    pos_lst = np.array(pos_lst)
    if len(pos_lst) > 1:
        d = pdist(pos_lst)
        return np.min(d)
    else:
        return 0.5


def _angles_in_ellipse(num, a, b):
    """Calculate angles of evenly spaced points in an ellipse

    Based on https://stackoverflow.com/a/52062369, 
    which is from https://pypi.org/project/flyingcircus/

    Parameters
    ----------
    num : int
        Sample number to get
    a : float
        Length of shorter axis
    b : float
        Length of longer axis

    Returns
    -------
    angles : ndarray
        Angles of sampled points
    """
    assert num > 0
    assert a < b
    angles = 2 * np.pi * np.arange(num) / num
    if a != b:
        e2 = 1.0 - a**2.0 / b**2.0
        tot_size = sp.special.ellipeinc(2.0 * np.pi, e2)
        arc_size = tot_size / num
        arcs = np.arange(num) * arc_size
        res = sp.optimize.root(lambda x: (sp.special.ellipeinc(x, e2) - arcs), angles)
        angles = res.x
    return angles


def _get_pos_one_part(nodes_show):
    """Find positions of nodes on a circle

    Also provides the minimum distances between nodes.

    Parameters
    ----------
    nodes_show : list of str
        Node names

    Returns
    -------
    pos : dict
        The positions for each node
    d_min : float
        Minimum distances between nodes
    """
    # positions of nodes
    n = len(nodes_show)

    # generate positions
    angle = _angles_in_ellipse(n, 0.999, 1.0)
    cen = [0.0, 0]
    rad = [1.0, 1.0]

    pos = dict()
    d_min = _add_node_to_a_circle(pos, nodes_show, cen, rad, angle)
    return pos, d_min


def _get_pos_two_parts(nodes_show, nodes_type):
    """Find positions of nodes with two groups.

    Group 1 on the left ellipse, group 2 on the right.
    Also provides the minimum distances between nodes.

    Parameters
    ----------
    nodes_show : list of str
        Node names
    nodes_type : dict
        Group ID for each node

    Returns
    -------
    pos : dict
        The positions for each node
    d_min : float
        Minimum distances between nodes
    """
    # positions of nodes
    nodes_sp = []
    nodes_tf = []
    for node in nodes_show:
        if nodes_type[node] == 0:
            nodes_sp.append(node)
        if nodes_type[node] == 1:
            nodes_tf.append(node)

    n_sp = len(nodes_sp)
    n_tf = len(nodes_tf)

    # generate positions
    angle_sp = _angles_in_ellipse(n_sp, 0.4, 1.0)
    angle_tf = _angles_in_ellipse(n_tf, 0.4, 1.0)

    cen_sp = [-0.6, 0]
    cen_tf = [0.6, 0]
    rad_sp = [0.4, 1]
    rad_tf = [0.4, 1]

    pos = dict()
    d_min_sp = _add_node_to_a_circle(pos, nodes_sp, cen_sp, rad_sp, angle_sp)
    d_min_tf = _add_node_to_a_circle(pos, nodes_tf, cen_tf, rad_tf, angle_tf)
    d_min = min(d_min_sp, d_min_tf)
    return pos, d_min


def _create_nx_graph(
    edges_df,
    nodes_show,
    min_alpha=0.2,
    max_alpha=1.0,
    mode="common",
    nodes_type=None,
):
    """Create NetworkX graph based on edge data frame.

    Add nodes and edges. Provide visualization related properties to the nodes.

    For common network, if two nodes in an edge have same type, draw grey line.
    If  two nodes in an edge have different type, draw green line.

    For differential network, if an edge comes from condition 1, draw blue line.
    If an edge comes from condition 2, draw red line.

    Parameters
    ----------
    edges_df : pandas.DataFrame
        Edge information. 
        First two columns for the two feature names. 
        Third column for edge type (common=0, diff1=1, diff2=2)
        Fourth column for weight.
    nodes_show : list of str
        Name of nodes to draw
    min_alpha : float, optional
        Minimum alpha value of edges, by default 0.2
        This is for the most light edges.
    max_alpha : float, optional
        Maximum alpha value of edges, by default 1.0
    mode : str, optional
        Draw common graph or differential graph, by default "common"
    nodes_type : None or dict, optional
        Node type (e.g., Gene=0, TF=1), by default None

    Returns
    -------
    nx.Graph
        Generated graph
    """
    # create the overall graph
    color_condition = {0: [0.7, 0.7, 0.7], 1: [0, 0, 1], 2: [1, 0, 0], 3: [0, 0.6, 0.3]}
    beta_max = np.max(edges_df["weight"])

    if nodes_type is None:
        nodes_type = dict()
        for node in nodes_show:
            nodes_type[node] = 0

    G = nx.Graph()
    for node in nodes_show:
        G.add_node(node)

    for i in range(len(edges_df)):
        gene1, gene2, condition, beta = edges_df.iloc[i]
        if condition in color_condition:
            alpha = np.abs(beta) / beta_max * (max_alpha - min_alpha) + min_alpha
            weight = np.abs(beta) / beta_max * 3.0 + 0.5
            if mode != "common":
                color = list(1 - (1 - np.array(color_condition[condition])) * alpha)
            else:
                if nodes_type[gene1] == nodes_type[gene2]:
                    color = list(1 - (1 - np.array(color_condition[0])) * alpha)
                else:
                    color = list(1 - (1 - np.array(color_condition[3])) * alpha)
            G.add_edge(gene1, gene2, color=color, weight=weight)

    return G


def _plot_network_helper(
    G,
    pos,
    d_min,
    labels,
    fig_size=18,
    font_size_scale=1,
    node_size_scale=2,
):
    """Draw the network

    The graph elements are optimized for figure size 18 x 18.
    For smaller figures, users may need to change `font_size_scale` and `node_size_scale`.

    Parameters
    ----------
    G : nx.Graph
        Graph to draw
    pos : dict
        Position of each node
    d_min : float
        Minimum distance between nodes.
        We use this to adjust node size, font size, etc.
    labels : dict
        Alternative names for nodes
    fig_size : int, optional
        Size of figure, by default 18
        We draw in a square figure, so width=height=`fig_size`.
    font_size_scale : int, optional
        Scale of fonts, by default 1
        To make the font larger, set larger value.
        The default value is scaled according to the node number.
    node_size_scale : int, optional
        Scale of node sizes, by default 2
    """
    # scale according to distances between nodes
    s_min = (d_min * 100) ** 2
    s_min = min(s_min, 500)

    # node size
    node_size = np.array([d for n, d in G.degree()])
    node_size = node_size / (np.max(node_size) + 1)
    node_size = node_size * s_min * node_size_scale * 10

    # font size
    font_size = d_min * font_size_scale * 100
    font_size = min(font_size, 10)

    # edges properties
    edges = G.edges()
    edge_color = [G[u][v]["color"] for u, v in edges]
    edge_weight = np.array([G[u][v]["weight"] for u, v in edges])

    # when there are too many edges, make the edge thin
    if len(edge_weight) > 400:
        edge_weight = edge_weight / len(edge_weight) * 400

    # draw
    fig, ax = plt.subplots(figsize=(fig_size, fig_size))

    nx.draw_networkx_nodes(
        G,
        pos=pos,
        ax=ax,
        node_color="lightblue",
        node_size=node_size,
        alpha=0.5,
    )

    nx.draw_networkx_edges(
        G,
        pos=pos,
        ax=ax,
        edgelist=edges,
        edge_color=edge_color,
        width=edge_weight,
    )

    nx.draw_networkx_labels(
        G,
        pos=pos,
        ax=ax,
        labels=labels,
        font_size=font_size,
        font_color="blueviolet",
    )

    ax.set_xlim((-1.1, 1.1))
    ax.set_ylim(ax.get_xlim())



"""The DDN main functions

There are two top level DDN functions: the `ddn` and the `ddn_parallel`.
The former one is in serial, which is easier to debug, while the latter one is in parallel.
For smaller data sets (e.g., less than 500 nodes), the serial version is fast enough.
The two functions should have the same functionality.

Each function allow using several different methods.

- `org`: the method in DDN 2.0. This is slow for larger data sets.
- `resi`: the method in DDN 3.0 using residual update strategy. This is suitable for larger feature number.
- `corr`: the method in DDN 3.0 using correlation matrix update strategy. This is suitable for larger sample number.
- `strongrule`: using the strong rule to accelerate the condition when lambda=0.

We recommend using `resi` in general case. In case you have much more samples than features, consider `corr`.

The choice of two hyperparameters lambda1 and lambda2 is critical.
We recommend running DDN with a range of parameters, and using prior knowledge to select the suitable.
Alternatively, users may refer the parameter tuning tutorial for other methods of choosing the parameters.

These functions also support using warm start.
To use this function, input the coefficient matrix from the previous call of DDN.
However, we do not observe very significant speed up, and are not generally recommended.

"""

import numpy as np
import joblib
from joblib import Parallel, delayed
from ddn3 import tools, solver


def ddn_parallel(
    g1_data,
    g2_data,
    lambda1=0.30,
    lambda2=0.10,
    threshold=1e-6,
    mthd="resi",
    std_est="std",
    g_rec_in=(),
    n_process=1,
):
    """Run DDN in parallel.

    Denote P be the number features. N1 be the sample size for condition 1, and N2 for condition 2.

    Parameters
    ----------
    g1_data : array_like, shape N1 by P
        The data from condition 1
    g2_data : array_like, shape N2 by P
        The data from condition 2
    lambda1 : float
        DDN parameter lambda1.
    lambda2 : float
        Not used. Must be 0.
    threshold : float
        Convergence threshold.
    mthd : str
        The DDN solver to use.
    std_est : str
        The standardization method. The default value is fine in most cases.
    g_rec_in : ndarray
        The input coefficient matrix. If warm start is not needed, do not change it.
    n_process : int
        Number of cores to use. Do not exceed the number of cores in your computer.
        If set to 1, no parallelization is used.

    Returns
    -------
    g_res : ndarray
        The estimated coefficient array of shape (2, P, P).
        g_res[0] is for the first condition, and g_res[1] for the second condition.

    """
    if n_process <= 1:
        n_process = int(joblib.cpu_count() / 2)

    n_node = g1_data.shape[1]
    n1 = g1_data.shape[0]
    n2 = g2_data.shape[0]
    g1_data = tools.standardize_data(g1_data, scaler=std_est)
    g2_data = tools.standardize_data(g2_data, scaler=std_est)

    if len(g_rec_in) == 0:
        g_rec_in = np.zeros((2, n_node, n_node))
        use_warm = False
    else:
        use_warm = True

    if mthd == "corr":
        corr_matrix_1 = g1_data.T @ g1_data / n1
        corr_matrix_2 = g2_data.T @ g2_data / n2
    else:
        corr_matrix_1 = []
        corr_matrix_2 = []

    if mthd == "org":
        out = Parallel(n_jobs=n_process)(
            delayed(solver.run_org)(
                g1_data,
                g2_data,
                node,
                lambda1,
                lambda2,
                beta1_in=g_rec_in[0][node],
                beta2_in=g_rec_in[1][node],
                threshold=threshold,
                use_warm=use_warm,
            )
            for node in range(n_node)
        )
    elif mthd == "resi":
        out = Parallel(n_jobs=n_process)(
            delayed(solver.run_resi)(
                g1_data,
                g2_data,
                node,
                lambda1,
                lambda2,
                beta1_in=g_rec_in[0][node],
                beta2_in=g_rec_in[1][node],
                threshold=threshold,
                use_warm=use_warm,
            )
            for node in range(n_node)
        )
    elif mthd == "corr":
        out = Parallel(n_jobs=n_process)(
            delayed(solver.run_corr)(
                corr_matrix_1,
                corr_matrix_2,
                node,
                lambda1,
                lambda2,
                beta1_in=g_rec_in[0][node],
                beta2_in=g_rec_in[1][node],
                threshold=threshold,
                use_warm=use_warm,
            )
            for node in range(n_node)
        )
    elif mthd == "strongrule":
        out = Parallel(n_jobs=n_process)(
            delayed(solver.run_strongrule)(
                g1_data,
                g2_data,
                node,
                lambda1,
                lambda2,
                beta1_in=g_rec_in[0][node],
                beta2_in=g_rec_in[1][node],
                threshold=threshold,
                use_warm=use_warm,
            )
            for node in range(n_node)
        )
    else:
        raise ("Method not implemented")

    g_rec = np.zeros((2, n_node, n_node))
    for node in range(n_node):
        g_rec[0, node, :] = out[node][0]
        g_rec[1, node, :] = out[node][1]

    return g_rec


def ddn(
    g1_data,
    g2_data,
    lambda1=0.30,
    lambda2=0.10,
    threshold=1e-6,
    mthd="resi",
    std_est="std",
    g_rec_in=(),
):
    """Run DDN.

    Denote P be the number features. N1 be the sample size for condition 1, and N2 for condition 2.

    Parameters
    ----------
    g1_data : array_like, shape N1 by P
        The data from condition 1
    g2_data : array_like, shape N2 by P
        The data from condition 2
    lambda1 : float
        DDN parameter lambda1.
    lambda2 : float
        Not used. Must be 0.
    threshold : float
        Convergence threshold.
    mthd : str
        The DDN solver to use.
    std_est : str
        The standardization method. The default value is fine in most cases.
    g_rec_in : ndarray
        The input coefficient matrix. If warm start is not needed, do not change it.

    Returns
    -------
    g_res : ndarray
        The estimated coefficient array of shape (2, P, P).
        g_res[0] is for the first condition, and g_res[1] for the second condition.

    """
    n_node = g1_data.shape[1]
    n1 = g1_data.shape[0]
    n2 = g2_data.shape[0]
    g1_data = tools.standardize_data(g1_data, scaler=std_est)
    g2_data = tools.standardize_data(g2_data, scaler=std_est)

    if len(g_rec_in) == 0:
        g_rec_in = np.zeros((2, n_node, n_node))
        use_warm = False
    else:
        use_warm = True

    if mthd == "corr":
        corr_matrix_1 = g1_data.T @ g1_data / n1
        corr_matrix_2 = g2_data.T @ g2_data / n2
    else:
        corr_matrix_1 = []
        corr_matrix_2 = []

    g_rec = np.zeros((2, n_node, n_node))
    for node in range(n_node):
        beta1_in = g_rec_in[0][node]
        beta2_in = g_rec_in[1][node]

        if mthd == "org":
            beta1, beta2 = solver.run_org(
                g1_data,
                g2_data,
                node,
                lambda1,
                lambda2,
                beta1_in,
                beta2_in,
                threshold,
                use_warm,
            )
        elif mthd == "resi":
            beta1, beta2 = solver.run_resi(
                g1_data,
                g2_data,
                node,
                lambda1,
                lambda2,
                beta1_in,
                beta2_in,
                threshold,
                use_warm,
            )
        elif mthd == "corr":
            beta1, beta2 = solver.run_corr(
                corr_matrix_1,
                corr_matrix_2,
                node,
                lambda1,
                lambda2,
                beta1_in,
                beta2_in,
                threshold,
                use_warm,
            )
        elif mthd == "strongrule":
            beta1, beta2 = solver.run_strongrule(
                g1_data,
                g2_data,
                node,
                lambda1,
                lambda2,
                beta1_in,
                beta2_in,
                threshold,
                use_warm,
            )
        else:
            print("Method not implemented")
            break

        g_rec[0, node, :] = beta1
        g_rec[1, node, :] = beta2

    return g_rec



"""Visualization functions of DDN, improved v2 version

These functions are still quite basic.
For advanced plotting of networks, consider using specialized tools.
"""

import math
import numpy as np
import scipy as sp
from scipy.spatial.distance import pdist
import networkx as nx
import matplotlib.pyplot as plt


def draw_network_for_ddn(
    edges_df,
    nodes_to_draw,
    mode="common",
    nodes_type=None,
    cen_lst=None,
    rad_lst=None,
    labels=None,
    fig_size=(16, 9),
    font_size_scale=1,
    node_size_scale=1,
    pdf_name="",
):
    """Draw networks for DDN

    Support drawing any number of ellipses according to `nodes_type`.
    The positions and shapes are given in `cen_lst` and `rad_lst`.
    This makes the layout of the graphs more flexible.

    The direction of the labels now points to the center of each ellipse.
    This allows showing larger fonts.

    The node size, font size, edge weight are now automatically adjusted according to figure size and node number.

    For common network, if two nodes in an edge have same type, draw grey line.
    If two nodes in an edge have different type, draw green line.

    For differential network, if an edge comes from condition 1, draw blue line.
    If an edge comes from condition 2, draw red line.

    Parameters
    ----------
    edges_df : pandas.DataFrame
        Edge information.
        First two columns for the two feature names.
        Third column for edge type (common=0, diff1=1, diff2=2)
        Fourth column for weight.
    nodes_to_draw : list of str
        Name of nodes to draw
    fig_size : tuple, optional
        Size of figure.
    font_size_scale : int, optional
        Scale of fonts, by default 1
        To make the font larger, set larger value.
        The default value is scaled according to the node number.
    node_size_scale : int, optional
        Scale of node sizes, by default 2
    nodes_type : None or dict, optional
        Node type (e.g., Gene=0, TF=1), by default None
    cen_lst : None or ndarray, optional
        The center of ellipse for each type of node. For node type i, use cen_lst[i]
        Shape (k, 2), k is the number of types. 2 for (x, y)
        Assume the y-axis of the figure is in range [-1,1], so do not set too large values.
    rad_lst : None or ndarray, optional
        The radius of ellipse for each type of node. For node type i, use rad_lst[i]
        Shape (k, 2), k is the number of types. 2 for shorter and longer axis length.
        Assume the y-axis of the figure is in range [-1,1], so do not set too large values.
    labels : dict
        Alternative (e.g., simplified) names for nodes
    mode : str, optional
        Draw common graph or differential graph, by default "common"
    pdf_name : str, optional
        Name of the PDF file to export, by default "". If set as "", no pdf will be output.

    Returns
    -------
    nx.Graph
        Graph object
    """
    # by default, assume there is one node type, and we draw circle
    if nodes_type is None:
        nodes_type = dict()
        for node in nodes_to_draw:
            nodes_type[node] = 0

    if cen_lst is None:
        cen_lst = np.array([[0.0, 0.0]])

    if rad_lst is None:
        rad_lst = np.array([[1.0, 1.0]])

    # create networkx graph
    G = create_nx_graph(
        edges_df,
        nodes_to_draw,
        mode=mode,
        nodes_type=nodes_type,
    )

    if labels is None:
        labels = dict((n, n) for n in G.nodes())

    # nodes positions
    pos, d_min = get_pos_multi_parts(
        nodes_to_draw, nodes_type, cen_lst=cen_lst, rad_lst=rad_lst
    )

    # plot the network
    plot_network(
        G,
        pos,
        d_min=d_min,
        labels=labels,
        node_type=nodes_type,
        cen_lst=cen_lst,
        fig_size=fig_size,
        font_size_scale=font_size_scale,
        node_size_scale=node_size_scale,
    )

    # export figure
    if len(pdf_name) > 0:
        plt.savefig(f"{pdf_name}_{mode}.pdf", format="pdf", bbox_inches="tight")

    return G


def create_nx_graph(
    edges_df,
    nodes_show,
    min_alpha=0.2,
    max_alpha=1.0,
    mode="common",
    nodes_type=None,
):
    """Create NetworkX graph based on edge data frame.

    Add nodes and edges. Provide visualization related properties to the nodes.

    For common network, if two nodes in an edge have same type, draw grey line.
    If  two nodes in an edge have different type, draw green line.

    For differential network, if an edge comes from condition 1, draw blue line.
    If an edge comes from condition 2, draw red line.

    Parameters
    ----------
    edges_df : pandas.DataFrame
        Edge information.
        First two columns for the two feature names.
        Third column for edge type (common=0, diff1=1, diff2=2)
        Fourth column for weight.
    nodes_show : list of str
        Name of nodes to draw
    min_alpha : float, optional
        Minimum alpha value of edges, by default 0.2
        This is for the most light edges.
    max_alpha : float, optional
        Maximum alpha value of edges, by default 1.0
    mode : str, optional
        Draw common graph or differential graph, by default "common"
    nodes_type : None or dict, optional
        Node type (e.g., Gene=0, TF=1), by default None

    Returns
    -------
    nx.Graph
        Generated graph
    """
    # create the overall graph
    color_condition = {0: [0.7, 0.7, 0.7], 1: [0, 0, 1], 2: [1, 0, 0], 3: [0, 0.6, 0.3]}
    beta_max = np.max(edges_df["weight"])

    if nodes_type is None:
        nodes_type = dict()
        for node in nodes_show:
            nodes_type[node] = 0

    G = nx.Graph()
    for node in nodes_show:
        G.add_node(node)

    for i in range(len(edges_df)):
        gene1, gene2, condition, beta = edges_df.iloc[i]
        if condition in color_condition:
            alpha = np.abs(beta) / beta_max * (max_alpha - min_alpha) + min_alpha
            weight = np.abs(beta) / beta_max * 3.0 + 0.5
            if mode != "common":
                color = list(1 - (1 - np.array(color_condition[condition])) * alpha)
            else:
                if nodes_type[gene1] == nodes_type[gene2]:
                    color = list(1 - (1 - np.array(color_condition[0])) * alpha)
                else:
                    color = list(1 - (1 - np.array(color_condition[3])) * alpha)
            G.add_edge(gene1, gene2, color=color, weight=weight)

    return G


def plot_network(
    G,
    pos,
    d_min,
    labels,
    node_type,
    cen_lst,
    fig_size,
    font_size_scale=1,
    node_size_scale=2,
):
    """Draw the network

    Parameters
    ----------
    G : nx.Graph
        Graph to draw
    pos : dict
        Position of each node
    d_min : float
        Minimum distance between nodes.
        We use this to adjust node size, font size, etc.
    labels : dict
        Alternative names for nodes
    node_type : None or dict, optional
        Node type (e.g., Gene=0, TF=1), by default None
    cen_lst : None or ndarray, optional
        The center of ellipse for each type of node. For node type i, use cen_lst[i]
        Shape (k, 2), k is the number of types. 2 for (x, y)
        Assume the y-axis of the figure is in range [-1,1], so do not set too large values.
    fig_size : tuple
        Size of figure. The unit is inch.
    font_size_scale : int, optional
        Scale of fonts, by default 1
        To make the font larger, set larger value.
        The default value is scaled according to the node number.
    node_size_scale : int, optional
        Scale of node sizes, by default 1
    """
    # The positions are given in a [-a,a]x[-1,1] region
    # Re-scale it to figure size, but leave some margin for text (here 0.8)
    fig_half_size = fig_size[1] * 0.9 / 2
    for x in pos:
        pos[x] = pos[x] * fig_half_size
    cen_lst = cen_lst * fig_half_size
    d_min = d_min * fig_half_size

    # node size
    # Node size in unit points^2. 1 inch = 72 points.
    # in case all nodes have degree zero
    # too large nodes are ugly
    s_min = (d_min * 72) ** 2
    s_min = min(s_min, 36*36)
    node_size = np.array([d for n, d in G.degree()]) + 0.1
    node_size = node_size / (np.max(node_size) + 1)
    node_size = node_size * s_min * node_size_scale

    # font size
    # In points. 1 inch = 72 points. Font size about the height of a character.
    # too large font may go outside the figure
    font_size = d_min * 72 * 0.8 * font_size_scale
    font_size = min(font_size, fig_half_size * 0.1 * 20)
    font_size_lst = font_size + node_size*0
    # font_size_lst = font_size * (
    #     np.abs(node_size) / np.max(node_size) * (1.0 - 0.5) + 0.5
    # )
    font_alpha_lst = np.abs(node_size) / np.max(node_size) * (1.0 - 0.1) + 0.1

    # draw
    fig, ax = plt.subplots(figsize=fig_size)

    nx.draw_networkx_nodes(
        G,
        pos=pos,
        ax=ax,
        node_color="lightblue",
        node_size=node_size,
        alpha=0.5,
    )

    # edges properties
    # in case there are no edges
    # too thick edges are ugly
    edges = G.edges()
    if len(edges) > 0:
        edge_color = [G[u][v]["color"] for u, v in edges]
        edge_weight = np.array([G[u][v]["weight"] for u, v in edges])

        # when there are too many edges, make the edge thin
        # edge weight also in points, 1 inch = 72 points
        d_min1 = min(d_min, 0.5)
        edge_weight = edge_weight / np.max(edge_weight) * d_min1 * 72 / 6
        if len(edge_weight) > 200:
            edge_weight = edge_weight / len(edge_weight) * 200

        nx.draw_networkx_edges(
            G,
            pos=pos,
            ax=ax,
            edgelist=edges,
            edge_color=edge_color,
            width=edge_weight,
        )

    _draw_network_labels(
        ax, pos, d_min, node_type, cen_lst, labels, font_size_lst, font_alpha_lst
    )

    ax.set_xlim((-fig_size[0] / 2, fig_size[0] / 2))
    ax.set_ylim((-fig_size[1] / 2, fig_size[1] / 2))


def get_pos_multi_parts(
    nodes_all,
    nodes_type,
    cen_lst=np.array([[-0.6, 0], [0.6, 0]]),
    rad_lst=np.array([[0.4, 1], [0.4, 1]]),
):
    """Find positions of nodes with multiple isolated subgraphs.

    Each subgraph is an ellipse.
    Also provides the minimum distances between nodes.

    Parameters
    ----------
    nodes_all : list of str
        Node names
    nodes_type : dict
        feature type ID for each node
    cen_lst : None or ndarray, optional
        The center of ellipse for each type of node. For node type i, use cen_lst[i]
        Shape (k, 2), k is the number of types. 2 for (x, y)
        Assume the y-axis of the figure is in range [-1,1], so do not set too large values.
    rad_lst : None or ndarray, optional
        The radius of ellipse for each type of node. For node type i, use rad_lst[i]
        Shape (k, 2), k is the number of types. 2 for shorter and longer axis length.
        Assume the y-axis of the figure is in range [-1,1], so do not set too large values.

    Returns
    -------
    pos : dict
        The positions for each node
    d_min : float
        Minimum distances between nodes
    """
    # types of features
    fea_type_lst = []
    for node in nodes_type:
        fea_type_lst.append(nodes_type[node])
    fea_type_enu = np.unique(np.array(fea_type_lst))

    # create positions for each type of features
    pos = dict()
    d_min_lst = []
    cntt = 0
    for fea_id in fea_type_enu:
        nodes = []
        for node in nodes_all:
            if nodes_type[node] == fea_id:
                nodes.append(node)
        pos, d_min = _get_pos_one_part(
            nodes, cen=cen_lst[cntt], rad=rad_lst[cntt], pos=pos
        )
        d_min_lst.append(d_min)
        cntt += 1
    d_min = np.min(np.array(d_min_lst))

    return pos, d_min


def _get_pos_one_part(
    nodes_show,
    cen=(0.0, 0.0),
    rad=(1.0, 1.0),
    pos=None,
):
    """Find positions of nodes on a circle

    Also provides the minimum distances between nodes.

    Parameters
    ----------
    nodes_show : tuple of str
        Node names
    cen : ndarray, optional
        The center of ellipse for each type of node. Shape (2, )
        Assume the y-axis of the figure is in range [-1,1], so do not set too large values.
    rad : ndarray, optional
        The radius of ellipse for each type of node. Shape (2, )
        Assume the y-axis of the figure is in range [-1,1], so do not set too large values.
    pos : dict
        NetworkX position dictionary

    Returns
    -------
    pos : dict
        The positions for each node
    d_min : float
        Minimum distances between nodes
    """
    n = len(nodes_show)
    r0 = float(rad[0])
    r1 = float(rad[1])
    if r0 == r1:
        r0 = r0 * 0.99999
    if r0 > r1:
        r0, r1 = r1, r0
    angle = _angles_in_ellipse(n, r0, r1)

    if pos is None:
        pos = dict()
    d_min = _add_node_to_a_circle(pos, nodes_show, cen, rad, angle)
    return pos, d_min


def _draw_network_labels(
    ax,
    pos,
    d_min,
    node_type,
    cen_lst,
    labels,
    font_size_lst,
    font_alpha_lst,
):
    """Add labels to a graph

    Modified from Networkx add label function

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes of the figure
    pos : dict
        NetworkX position dictionary
    d_min : float
        Minimum distance between nodes
    node_type : dict
        Feature type index of each node
    cen_lst : None or ndarray, optional
        The center of ellipse for each type of node. For node type i, use cen_lst[i]
        Shape (k, 2), k is the number of types. 2 for (x, y)
        Assume the y-axis of the figure is in range [-1,1], so do not set too large values.
    labels : dict
        Alternative names for each node, to be shown in the graph
    font_size_lst : array_like
        Font size for each node, in points
    font_alpha_lst : array_like
        Font alpha for each node, value from 0 to 1
    """
    cnt = 0
    for node, label in labels.items():
        (x, y) = pos[node]
        fea_idx = node_type[node]
        x1 = x - cen_lst[fea_idx][0]
        y1 = y - cen_lst[fea_idx][1]

        # rotate labels to make it point to the center
        # this allows larger fonts
        # treat labels on the right and those on the left differently
        angle = math.atan2(y1, np.abs(x1)) / math.pi * 180
        if x1 < 0:
            angle = -angle

        # this makes "1" and 1 labeled the same
        if not isinstance(label, str):
            label = str(label)

        nn = np.sqrt(x1**2 + y1**2)
        x1n = x1 / nn
        y1n = y1 / nn

        x_ext = x + x1n * (len(label) / 2 * font_size_lst[cnt] / 72 + d_min / 4)
        y_ext = y + y1n * (len(label) / 2 * font_size_lst[cnt] / 72 + d_min / 4)

        _ = ax.text(
            x_ext,
            y_ext,
            label,
            size=font_size_lst[cnt],
            color="k",
            family="sans-serif",
            weight="normal",
            alpha=font_alpha_lst[cnt],
            horizontalalignment="center",
            verticalalignment="center",
            transform=ax.transData,
            rotation=angle,
            bbox=None,
            clip_on=True,
        )
        cnt += 1

    ax.tick_params(
        axis="both",
        which="both",
        bottom=False,
        left=False,
        labelbottom=False,
        labelleft=False,
    )


def _add_node_to_a_circle(pos, nodes, cen, rad, angles):
    """Find positions in an ellipse, and calculate the minimum distances between points

    Parameters
    ----------
    pos : dict
        For saving the position results for NetworkX
    nodes : list of str
        Name of nodes
    cen : array_like
        Central position of this circle, shape (2,)
    rad : array_like
        Length of two axes of the ellipse, shape (2,)
    angles : array_like
        Angles of the points

    Returns
    -------
    float
        Minimum distances between points
    """
    pos_lst = []
    cnt = 0
    for node in nodes:
        theta = angles[cnt]
        pos0 = np.array(
            [cen[0] + np.cos(theta) * rad[0], cen[1] + np.sin(theta) * rad[1]]
        )
        pos[node] = pos0
        pos_lst.append(pos0)
        cnt += 1
    pos_lst = np.array(pos_lst)
    if len(pos_lst) > 1:
        d = pdist(pos_lst)
        return np.min(d)
    else:
        return 0.5


def _angles_in_ellipse(num, a, b):
    """Calculate angles of evenly spaced points in an ellipse

    Based on https://stackoverflow.com/a/52062369,
    which is from https://pypi.org/project/flyingcircus/

    Parameters
    ----------
    num : int
        Sample number to get
    a : float
        Length of shorter axis
    b : float
        Length of longer axis

    Returns
    -------
    angles : ndarray
        Angles of sampled points
    """
    assert num > 0
    assert a < b
    angles = 2 * np.pi * np.arange(num) / num
    if a != b:
        e2 = 1.0 - a**2.0 / b**2.0
        tot_size = sp.special.ellipeinc(2.0 * np.pi, e2)
        arc_size = tot_size / num
        arcs = np.arange(num) * arc_size
        res = sp.optimize.root(lambda x: (sp.special.ellipeinc(x, e2) - arcs), angles)
        angles = res.x
    return angles



"""Functions that implements the strong rule strategy in DDN 3.0
"""

import numpy as np
from sklearn.linear_model import Lasso


def strong_rule(X, y, lambda1):
    """Find the variables to keep using the strong rule

    This rule eliminates a set of variables that will generally not be selected by Lasso.
    Let N be the sample size. Let P be the feature size.

    Parameters
    ----------
    X : array_like
        The predictors. Shape N by P-1.
    y : array_like
        The response variables. Shape N by 1.
    lambda1 : float
        DDN parameter lambda1.

    Returns
    -------
    idx : ndarray
        List of indices to keep

    """
    n = len(y)
    lambdas = np.abs(np.matmul(X.T, y)) / n
    lambdamax = max(lambdas)
    idx = np.array(
        [i for i in range(X.shape[1]) if lambdas[i] >= 2 * lambda1 - lambdamax]
    )
    # drop = X.shape[1] - len(idx)
    return idx


def lasso(y, X, lambda1, beta_in, tol=1e-6, use_strong_rule=True, use_warm=False):
    """Compute lasso with strong rule that eliminates variables

    Let N be the sample size. Let P be the feature size.

    Parameters
    ----------
    y : array-like
        The response variable. Size N by 1.
    X : array_like
        The predictor. Size N by P-1
    lambda1 : float
        DDN parameter lambda1.
    beta_in : array_like
        Initial lasso coefficient. Size P-1 by 1.
    tol : float
        Lasso tolerance.
    use_strong_rule : bool
        Apply strong rule or not. Strong rule will eliminate some predictors.
    use_warm : bool
        Apply warm start or not.

    Returns
    -------
    beta : ndarray
        The estimated lasso coefficients. Size P-1.

    """
    # select feature index
    n_pred = X.shape[1]
    idx = np.arange(n_pred)

    if use_strong_rule:
        idx = strong_rule(X, y, lambda1)
        if len(idx) == 0:
            return np.zeros(n_pred)
        X = X[:, idx]

    # calculate beta with sklearn lasso
    if use_warm:
        clf = Lasso(
            alpha=lambda1,
            # max_iter=1000,
            tol=tol,
            fit_intercept=False,
            warm_start=True,
        )
        clf.coef_ = beta_in
    else:
        clf = Lasso(
            alpha=lambda1,
            # max_iter=1000,
            tol=tol,
            fit_intercept=False,
            warm_start=False,
        )
        clf.coef_ = np.zeros((n_pred,))
    clf.fit(X, y)

    beta = np.zeros(n_pred)
    # beta = np.zeros((len(idx),))
    beta[idx] = clf.coef_

    return beta



"""Tools for extrating information from DDN outputs
"""

import numpy as np
import pandas as pd
from ddn3 import tools


def get_diff_comm_net_for_plot(omega1, omega2, gene_names):
    """Generate common and differential networks from DDN outputs

    Parameters
    ----------
    omega1 : ndarray
        DDN estimated coefficient matrix for condition 1
    omega2 : ndarray
        DDN estimated coefficient matrix for condition 2
    gene_names : list of str
        Gene names

    Returns
    -------
    comm_edge : pandas.DataFrame
        Data frame containing information of edges in common networks.
        The first two columns means the two genes in each edge.
        The fourth column gives the weights of that edge. 
        The weights is the absolute value of the DDN coefficient.
    dif1_edge : pandas.DataFrame
        Data frame containing information of differential edges from condition 1.
        The contents are similart to `comm_edge`.
    dif2_edge : pandas.DataFrame
        Data frame containing information of differential edges from condition 2.
        The contents are similart to `comm_edge`.
    diff_edge : pandas.DataFrame
        Data frame containing information of edges in the differential networks.
        The contents are similart to `comm_edge`, except that the third column 
        labels whether this edge come from condition 1 or 2.
    node_non_isolated : list of str
        Node names that are involved in edges in either common network or differential networks.
    """
    
    # make the coefficient matrices symmetric
    omega1 = (omega1 + omega1.T) / 2
    omega2 = (omega2 + omega2.T) / 2
    omega_comm = (omega1 + omega2) / 2

    # apply a small threshold to obtain adjacency matrices
    g1 = tools.get_net_topo_from_mat(omega1)
    g2 = tools.get_net_topo_from_mat(omega2)
    gene_names = np.array(gene_names)

    # find common and differential networks
    comm_adj_mat = (g1 + g2) == 2
    dif1_adj_mat = g1 != comm_adj_mat
    dif2_adj_mat = g2 != comm_adj_mat

    # Get edges in format `gene1`, `gene2`, `condition`, `beta`
    comm_edge = _get_edge_list(comm_adj_mat, omega_comm, gene_names, group_idx=0)
    dif1_edge = _get_edge_list(dif1_adj_mat, omega1, gene_names, group_idx=1)
    dif2_edge = _get_edge_list(dif2_adj_mat, omega2, gene_names, group_idx=2)

    # find nodes that are involved in at least one edge
    node_degree_nz = (np.sum(g1, axis=0) > 0) + (np.sum(g2, axis=0) > 0)
    node_idx_non_isolated = np.sort(np.where(node_degree_nz > 0)[0])
    node_non_isolated = gene_names[node_idx_non_isolated]
    diff_edge = pd.concat((dif1_edge, dif2_edge))

    return comm_edge, dif1_edge, dif2_edge, diff_edge, node_non_isolated


def _get_edge_list(conn_mat, beta_mat, gene_names, group_idx=0):
    """Create a pandas data frame reporting the edges

    Parameters
    ----------
    conn_mat : ndarray
        Adjacency matrix for a graph
    beta_mat : ndarray
        The coefficient matrix of a graph
    gene_names : list of str
        Gane names
    group_idx : int, optional
        The group labeling, by default 0
        0 for common edges, 1 for condition 1, 2 for condition 2

    Returns
    -------
    pandas.DataFrame
        Data frame containing information of edges.
        The first two columns give names of genes for each edge.
        The third column is for the group label.
        The fourth column is the absolute value of the coefficient.
    """
    conn_mat1 = np.tril(conn_mat, -1)
    n1, n2 = np.where(conn_mat1 > 0)
    edge_weight = np.abs(beta_mat[n1, n2])
    out = dict(
        gene1=gene_names[n1],
        gene2=gene_names[n2],
        condition=group_idx,
        weight=edge_weight,
    )
    return pd.DataFrame(data=out)


def get_node_type_and_label_two_parts(
    nodes_show, part1_id="SP", part2_id="TF", ofst1=2, ofst2=1
):
    """Given a list of features, divide them to to groups, and clean their names.

    The first several characters in the name of each feature is related to their group information.
    For example, for feature "SP_AAAA", it means a gene in the "SP" group and the gene name is "AAAA".
    For feature "TF_BBBB", it means a gene in "TF" group with name "BBBB".
    We want to put thing starting with "TF" and "SP" to two separate groups.
    We also want to create a simplifed label that do not contain "SP" and "TP" to make plotting easier.
    
    Parameters
    ----------
    nodes_show : list of str
        The list of feature names
    part1_id : str, optional
        The label for group 1, by default "SP"
    part2_id : str, optional
        The label for group 2, by default "TF"
    ofst1 : int, optional
        The number of extra characters to remove the prefix before the gene name, by default 2
    ofst2 : int, optional
        The number of extra characters to remove the prefix before the gene name, by default 1

    Returns
    -------
    nodes_type : dict
        For each name in the feature, give its group index.
    labels : dict
        For each name in the feature, give its simplified name.
    """
    # TODO: support more parts
    x_len1 = len(part1_id)
    x_len2 = len(part2_id)

    # make labels shorter
    nodes_type = dict()
    labels = dict()
    for i, node in enumerate(nodes_show):
        if node[:x_len1] == part1_id:
            labels[node] = node[x_len1 + ofst1 :]
            nodes_type[node] = 0
        if node[:x_len2] == part2_id:
            labels[node] = node[x_len2 + ofst2 :]
            nodes_type[node] = 1

    return nodes_type, labels


def get_edge_subset_by_name(edge_df, name_match="TF"):
    """Extract subset of edges in the data frame.

    Parameters
    ----------
    edge_df : pandas.DataFrame
        The data frame containing edges.
    name_match : str, optional
        The pattern of the name to match, by default "TF"
        Must be at the beginning of the name of each feature.

    Returns
    -------
    pandas.DataFrame
        Extracted edges
    """
    gene_sel = np.zeros(len(edge_df))
    nn = len(name_match)
    for i in range(len(edge_df)):
        gene1, gene2, _, _ = edge_df.iloc[i]
        if gene1[:nn] == name_match and gene2[:nn] == name_match:
            gene_sel[i] = 1
    return edge_df[gene_sel > 0]


def get_node_subset_by_name(nodes, name_match="TF"):
    """Get a subset of the nodes according to the given pattern"""
    nodes_sel = []
    nn = len(name_match)
    for node in nodes:
        if node[:nn] == name_match:
            nodes_sel.append(node)
    return nodes_sel



"""Block coordinate descent for DDN

This module implements the original block coordinate descent in [1], as well as various accelerated versions.
Most methods are also Numba accelerated, except for `bcd_org_old`, which is left for comparison purposes.

For data with lots of samples, consider using `bcd_corr`.
For data with lots of features, the `bcd_residual` is faster.

[1] Zhang, Bai, and Yue Wang. "Learning structural changes of Gaussian graphical models in controlled experiments." UAI (2010).
"""

import numpy as np
import numba


def bcd_org_old(
    beta_in, y, X, lambda1, lambda2, n1=1, n2=1, threshold=1e-6, max_iter=100000
):
    """The BCD algorithm in DDN 2.0.

    See https://arxiv.org/abs/1203.3532 for details
    Denote P be the number features. N1 be the sample size for condition 1, and N2 for condition 2.

    Parameters
    ----------
    beta_in : array_like, length 2P
        Initial beta. If initialization is not needed, use an array of all zeros
    y : array_like, size (N1+N2) by 1
        The samples for the node serving as response.
    X : array_like, size (N1+N2) by (P-2)
        The block diagonal matrix with all the predictive nodes. Each condition has P-1 nodes.
        The top left block has size N1(P-1), while the bottom right block has size N2(P-1).
        Other elements are all zeros.
    lambda1 : float
        DDN parameter lambda1.
    lambda2 : float
        DDN parameter lambda2.
    n1 : int
        Sample size for condition 1, N1
    n2 : int
        Sample size for condition 2, N2
    threshold : float
        Convergence threshold.
    max_iter : int
        Maximum number of iterations

    Returns
    -------
    beta : ndarray, shape 2P
        Estimated beta for two conditions on node CurrIdx
    r : int
        Number of iterations taken
    betaerr : float
        The error term

    """
    # total feature size must be even
    if X.shape[1] == 0 or X.shape[1] % 2 == 1:
        return []

    # feature size (gene size) for each group
    # x1 (control): feature 0 to p-1
    # x2 (case): feature p to 2*p-1
    p = X.shape[1] // 2

    # initialize beta
    beta = np.copy(beta_in)
    # beta = np.zeros(2 * p)

    if p == 1:
        rho1 = np.sum(y * X[:, 0]) / n1
        rho2 = np.sum(y * X[:, 1]) / n2

        beta2d = solve2d(rho1, rho2, lambda1, lambda2)
        beta[0] = beta2d[0]
        beta[1] = beta2d[1]

        return beta
    else:
        r = 0
        while True:
            beta_old = np.copy(beta)

            for k in range(p):
                x1 = X[:, k]
                x2 = X[:, k + p]
                r = r + 1

                idx = [i for i in range(2 * p) if i not in (k, k + p)]
                y_residual = y - np.dot(X[:, idx], beta[idx])

                rho1 = np.sum(y_residual * x1) / n1
                rho2 = np.sum(y_residual * x2) / n2

                if np.abs(rho1) > 1000 or np.abs(rho2) > 1000:
                    print(rho1, rho2)

                beta2d = solve2d(rho1, rho2, lambda1, lambda2)

                beta[k] = beta2d[0]
                beta[k + p] = beta2d[1]

            betaerr = np.mean(np.abs(beta - beta_old))
            if (betaerr < threshold) or (r > max_iter):
                break

        return beta, r, betaerr


@numba.njit
def bcd_org(
    beta_in, y, X, lambda1, lambda2, n1=1, n2=1, threshold=1e-6, max_iter=100000
):
    """The BCD algorithm in DDN 2.0, with Numba acceleration and some simplification.

    See https://arxiv.org/abs/1203.3532 for details
    Denote P be the number features. N1 be the sample size for condition 1, and N2 for condition 2.

    Parameters
    ----------
    beta_in : array_like, length 2P
        Initial beta. If initialization is not needed, use an array of all zeros
    y : array_like, size (N1+N2) by 1
        The samples for the node serving as response.
    X : array_like, size (N1+N2) by (P-2)
        The block diagonal matrix with all the predictive nodes. Each condition has P-1 nodes.
        The top left block has size N1(P-1), while the bottom right block has size N2(P-1).
        Other elements are all zeros.
    lambda1 : float
        DDN parameter lambda1.
    lambda2 : float
        DDN parameter lambda2.
    n1 : int
        Sample size for condition 1, N1
    n2 : int
        Sample size for condition 2, N2
    threshold : float
        Convergence threshold.
    max_iter : int
        Maximum number of iterations

    Returns
    -------
    beta : ndarray, shape 2P
        Estimated beta for two conditions on node CurrIdx
    r : int
        Number of iterations taken
    betaerr : float
        The error term

    """
    # feature size (gene size) for each group
    # x1 (control): feature 0 to p-1
    # x2 (case): feature p to 2*p-1
    p = X.shape[1] // 2

    # initialize beta
    beta = np.copy(beta_in)
    r = 0
    while True:
        beta_old = np.copy(beta)

        for k in range(p):
            x1 = X[:, k]
            x2 = X[:, k + p]
            r = r + 1

            beta_temp = np.copy(beta)
            beta_temp[k] = 0
            beta_temp[k + p] = 0
            y_residual = y - np.dot(X, beta_temp)

            rho1 = np.sum(y_residual * x1) / n1
            rho2 = np.sum(y_residual * x2) / n2

            beta2d = solve2d(rho1, rho2, lambda1, lambda2)

            beta[k] = beta2d[0]
            beta[k + p] = beta2d[1]

        betaerr = np.mean(np.abs(beta - beta_old))
        if (betaerr < threshold) or (r > max_iter):
            break

    return beta, r, betaerr


@numba.njit
def bcd_residual(
    beta_in,
    X1,
    X2,
    y1_resi,
    y2_resi,
    CurrIdx,
    lambda1,
    lambda2,
    threshold,
    max_iter=10000,
):
    """BCD algorithm for DDN using residual update strategy

    The algorithm allows warm start, which requires initial `beta_in`, `y1_resi`, and `y2_resi`.
    See `run_resi` on how to prepare these inputs.
    Denote P be the number features. N1 be the sample size for condition 1, and N2 for condition 2.

    Parameters
    ----------
    beta_in : array_like, length 2P
        Initial beta. If initialization is not needed, use an array of all zeros
    X1 : array_like, shape N1 by P
        The data from condition 1
    X2 : array_like, shape N2 by P
        The data from condition 2
    y1_resi : array_like, shape N1 by 1
        The initial residual signal for condition 1. If warm start is not used, it is column CurrIdx of X1.
    y2_resi : array_like, shape N2 by 1
        The initial residual signal for condition 2. If warm start is not used, it is column CurrIdx of X2.
    CurrIdx : int
        Index of the current node that serve as the response variable.
    lambda1 : float
        DDN parameter lambda1.
    lambda2 : float
        DDN parameter lambda2.
    threshold : float
        Convergence threshold.
    max_iter : int
        Maximum number of iterations

    Returns
    -------
    beta : ndarray, shape 2P
        Estimated beta for two conditions on node CurrIdx
    r : int
        Number of iterations taken
    betaerr : float
        The error term

    """
    p = X1.shape[1]
    n1 = X1.shape[0]
    n2 = X2.shape[0]
    # n_mean = (n1+n2/2) # BUG

    beta1 = np.copy(beta_in[:p])
    beta2 = np.copy(beta_in[p:])

    betaerr_array = np.zeros(2 * p)
    beta = np.zeros(2 * p)

    r = 0
    k_last = CurrIdx
    while True:
        beta1_old = np.copy(beta1)
        beta2_old = np.copy(beta2)

        for i in range(p):
            if i == CurrIdx:
                continue

            r = r + 1
            k = i

            y1_resi = y1_resi - beta1[k_last] * X1[:, k_last] + beta1[k] * X1[:, k]
            y2_resi = y2_resi - beta2[k_last] * X2[:, k_last] + beta2[k] * X2[:, k]
            rho1 = np.sum(y1_resi * X1[:, k]) / n1
            rho2 = np.sum(y2_resi * X2[:, k]) / n2
            # rho1 = np.sum(y1_resi * X1[:, k]) / n_mean # BUG
            # rho2 = np.sum(y2_resi * X2[:, k]) / n_mean # BUG

            beta2d = solve2d(rho1, rho2, lambda1, lambda2)
            beta1[k] = beta2d[0]
            beta2[k] = beta2d[1]

            k_last = k

        betaerr_array[:p] = beta1 - beta1_old
        betaerr_array[p:] = beta2 - beta2_old
        betaerr = np.mean(np.abs(betaerr_array))

        if (betaerr < threshold) or (r > max_iter):
            break

    beta[:p] = beta1
    beta[p:] = beta2
    return beta, r, betaerr


@numba.njit
def bcd_corr(
    beta_in,
    cur_node,
    lambda1,
    lambda2,
    corr_matrix_1,
    corr_matrix_2,
    threshold=1e-6,
    max_iter=100000,
):
    """BCD algorithm for DDN using correlation matrix update strategy

    This approach is more suitable for larger sample sizes.
    The algorithm allows warm start, which requires initial `beta_in`.
    Denote P be the number features. N1 be the sample size for condition 1, and N2 for condition 2.

    Parameters
    ----------
    beta_in : array_like, length 2P
        Initial beta. If initialization is not needed, use an array of all zeros
    cur_node : int
        Index of the current node that serve as the response variable.
    lambda1 : float
        DDN parameter lambda1.
    lambda2 : float
        DDN parameter lambda2.
    corr_matrix_1 : array_like, P by P
        Correlation matrix for condition 1
    corr_matrix_2 : array_like, P by P
        Correlation matrix for condition 2
    threshold : float
        Convergence threshold.
    max_iter : int
        Maximum number of iterations

    Returns
    -------
    beta : ndarray, shape 2P
        Estimated beta for two conditions on node CurrIdx
    r : int
        Number of iterations taken
    delta_beta : float
        The error term

    """
    p = int(len(beta_in) / 2)
    beta1 = np.copy(beta_in[:p])
    beta2 = np.copy(beta_in[p:])
    beta_dif = np.zeros(2 * p)
    delta_beta = 0.0

    r = 0
    for _ in range(max_iter):
        beta1_old = np.copy(beta1)
        beta2_old = np.copy(beta2)

        for k in range(p):
            if k == cur_node:
                continue
            r = r + 1

            betaBar1 = -beta1
            betaBar2 = -beta2
            betaBar1[k] = 0
            betaBar2[k] = 0
            betaBar1[cur_node] = 1
            betaBar2[cur_node] = 1

            rho1 = np.sum(betaBar1 * corr_matrix_1[:, k])
            rho2 = np.sum(betaBar2 * corr_matrix_2[:, k])

            beta2d = solve2d(rho1, rho2, lambda1, lambda2)
            beta1[k] = beta2d[0]
            beta2[k] = beta2d[1]

        beta_dif[:p] = beta1 - beta1_old
        beta_dif[p:] = beta2 - beta2_old
        delta_beta = np.mean(np.abs(beta_dif))

        if delta_beta < threshold:
            break

        # print(delta_beta)

    beta = np.zeros(2 * p)
    beta[:p] = beta1
    beta[p:] = beta2

    return beta, r, delta_beta


@numba.njit
def solve2d(rho1, rho2, lambda1, lambda2):
    """Optimize for two variables corresponding to one node

    The details can be found in https://arxiv.org/abs/1203.3532

    Parameters
    ----------
    rho1 : float
        The rho from data1
    rho2 : float
        The rho from data2
    lambda1 : float
        DDN parameter lambda1
    lambda2 : float
        DDN parameter lambda2

    Returns
    -------
    beta1 : float
        Optimal coefficient for data1
    beta2 : float
        Optimal coefficient for data2

    """
    # initialize output
    area_index = 0
    beta1 = 0
    beta2 = 0

    if (
        rho2 <= (rho1 + 2 * lambda2)
        and rho2 >= (rho2 - 2 * lambda2)
        and rho2 >= (2 * lambda1 - rho1)
    ):
        area_index = 1
        beta1 = (rho1 + rho2) / 2 - lambda1
        beta2 = (rho1 + rho2) / 2 - lambda1
    if rho2 > (rho1 + 2 * lambda2) and rho1 >= (lambda1 - lambda2):
        area_index = 2
        beta1 = rho1 - lambda1 + lambda2
        beta2 = rho2 - lambda1 - lambda2
    if (
        rho1 < (lambda1 - lambda2)
        and rho1 >= -(lambda1 + lambda2)
        and rho2 >= (lambda1 + lambda2)
    ):
        area_index = 3
        beta1 = 0
        beta2 = rho2 - lambda1 - lambda2
    if rho1 < -(lambda1 + lambda2) and rho2 >= (lambda1 + lambda2):
        area_index = 4
        beta1 = rho1 + lambda1 + lambda2
        beta2 = rho2 - lambda1 - lambda2
    if (
        rho1 < -(lambda1 + lambda2)
        and rho2 < (lambda1 + lambda2)
        and rho2 >= -(lambda1 + lambda2)
    ):
        area_index = 5
        beta1 = rho1 + lambda1 + lambda2
        beta2 = 0
    if rho2 < -(lambda1 - lambda2) and rho2 >= (rho1 + 2 * lambda2):
        area_index = 6
        beta1 = rho1 + lambda1 + lambda2
        beta2 = rho2 + lambda1 - lambda2
    if (
        rho2 >= (rho1 - 2 * lambda2)
        and rho2 < (rho1 + 2 * lambda2)
        and rho2 <= (-2 * lambda1 - rho1)
    ):
        area_index = 7
        beta1 = (rho1 + rho2) / 2 + lambda1
        beta2 = (rho1 + rho2) / 2 + lambda1
    if rho2 < (rho1 - 2 * lambda2) and rho1 <= -(lambda1 - lambda2):
        area_index = 8
        beta1 = rho1 + lambda1 - lambda2
        beta2 = rho2 + lambda1 + lambda2
    if (
        rho1 <= (lambda1 + lambda2)
        and rho1 >= -(lambda1 - lambda2)
        and rho2 <= -(lambda1 + lambda2)
    ):
        area_index = 9
        beta1 = 0
        beta2 = rho2 + lambda1 + lambda2
    if rho1 > (lambda1 + lambda2) and rho2 <= -(lambda1 + lambda2):
        area_index = 10
        beta1 = rho1 - lambda1 - lambda2
        beta2 = rho2 + lambda1 + lambda2
    if (
        rho2 > -(lambda1 + lambda2)
        and rho2 <= (lambda1 - lambda2)
        and rho1 >= (lambda1 + lambda2)
    ):
        area_index = 11
        beta1 = rho1 - lambda1 - lambda2
        beta2 = 0
    if rho2 > (lambda1 - lambda2) and rho2 < (rho1 - 2 * lambda2):
        area_index = 12
        beta1 = rho1 - lambda1 - lambda2
        beta2 = rho2 - lambda1 + lambda2

    return [beta1, beta2]
